/*
 * Copyright (C) 2004-2008 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */
/** \file ogg_afh.c para_server's ogg vorbis audio format handler */

#include <inttypes.h>
#include <ogg/ogg.h>
#include <vorbis/codec.h>
#include <vorbis/vorbisfile.h>

#include "para.h"
#include "afh.h"
#include "error.h"
#include "string.h"
#include "afs.h"
#include "server.h"

/** must be big enough to hold header */
#define CHUNK_SIZE 32768
static double chunk_time = 0.25;

/** describes a memory-mapped ogg vorbis file */
struct ogg_datasource {
	/** the memory mapping */
	char *map;
	/** this size of the mapping */
	off_t numbytes;
	/** the current position in the mapping */
	off_t fpos;
};

static size_t cb_read(void *buf, size_t size, size_t nmemb, void *datasource)
{
	struct ogg_datasource *ods = datasource;
	size_t copy = PARA_MIN(ods->numbytes - ods->fpos, size * nmemb),
		ret = copy / size;
	if (!ret)
		return 0;
	memcpy(buf, ods->map + ods->fpos, copy);
//	PARA_INFO_LOG("size: %zd, nmemb: %zd, ret: %zd\n", size, nmemb, ret);
	ods->fpos += ret * size;
	return ret;
}

static int cb_seek(void *datasource, ogg_int64_t offset,
		int whence)
{
	struct ogg_datasource *ods = datasource;
	switch (whence) {
	case SEEK_SET:
		if (offset >= 0 && offset <= ods->numbytes) {
			ods->fpos = offset;
			return 0;
		}
		errno = EINVAL;
		return -1;
		break;
	case SEEK_END:
		if (offset <= 0 && -offset <= ods->numbytes) {
			ods->fpos = ods->numbytes + offset;
			return 0;
		}
		errno = EINVAL;
		return -1;
		break;
	case SEEK_CUR:
		if ((offset >= 0 && offset + ods->fpos > ods->numbytes) ||
				(offset < 0 && offset + ods->fpos < 0)) {
			errno = EINVAL;
			return -1;
		}
		ods->fpos += offset;
		return 0;
	}
	errno = EINVAL;
	return -1;
}

/* don't do anything as vss still needs the open filehandle */
static int cb_close(__a_unused void *datasource)
{
	return 0;
}

static long cb_tell(void *datasource)
{
	struct ogg_datasource *ods = datasource;
	return (unsigned long)ods->fpos;
}

static int ogg_open_callbacks(void *datasource, OggVorbis_File *vf, ov_callbacks c)
{
	int ret = ov_open_callbacks(datasource, vf,
		NULL, /* no initial buffer */
		0, /* no initial bytes */
		c); /* the ov_open_callbacks */

	if (ret == OV_EREAD)
		return -E_OGG_READ;
	if (ret == OV_ENOTVORBIS)
		return -E_VORBIS;
	if (ret == OV_EVERSION)
		return -E_OGG_VERSION;
	if (ret == OV_EBADHEADER)
		return -E_OGG_BAD_HEADER;
	if (ret < 0)
		return -E_OGG_UNKNOWN_ERROR;
	return 1;

}

static int ogg_compute_header_len(char *map, size_t numbytes,
		struct afh_info *afhi)
{
	int ret;
	size_t len = PARA_MIN(numbytes, CHUNK_SIZE);
	int serial;
	char *buf;

	ogg_page page;
	ogg_packet packet;
	vorbis_comment vc;
	vorbis_info vi;
	ogg_stream_state *stream_in = para_malloc(sizeof(ogg_stream_state));
	ogg_stream_state *stream_out = para_malloc(sizeof(ogg_stream_state));
	ogg_sync_state *sync_in = para_malloc(sizeof(ogg_sync_state));

	ogg_sync_init(sync_in);
	vorbis_info_init(&vi);
	vorbis_comment_init(&vc);
	buf = ogg_sync_buffer(sync_in, (long)len);
	memcpy(buf, map, len);
	ogg_sync_wrote(sync_in, (long)len);
	ret = -E_SYNC_PAGEOUT;
	if (ogg_sync_pageout(sync_in, &page) <= 0) {
		free(stream_in);
		free(stream_out);
		goto err1;
	}
	serial = ogg_page_serialno(&page);
	ogg_stream_init(stream_in, serial);
	ogg_stream_init(stream_out, serial);
	ret = ogg_stream_pagein(stream_in, &page);
	if (ret < 0) {
		ret = -E_STREAM_PAGEIN;
		goto err2;
	}
	ret = ogg_stream_packetout(stream_in, &packet);
	if (ret != 1) {
		ret = -E_STREAM_PACKETOUT;
		goto err2;
	}
	ret = -E_VORBIS;
	if (vorbis_synthesis_headerin(&vi, &vc, &packet) < 0)
		goto err2;
	PARA_DEBUG_LOG("channels: %i, rate: %li\n", vi.channels, vi.rate);
	ogg_stream_packetin(stream_out, &packet);

	ret = ogg_sync_pageout(sync_in, &page);
	if (ret <= 0) {
		ret = -E_SYNC_PAGEOUT;
		goto err2;
	}
	ogg_stream_pagein(stream_in, &page);
	ogg_stream_packetout(stream_in, &packet);
	ogg_stream_packetin(stream_out, &packet);

	ret = ogg_sync_pageout(sync_in, &page);
	if (ret <= 0) {
		ret = -E_SYNC_PAGEOUT;
		goto err2;
	}
	ogg_stream_pagein(stream_in, &page);
	ogg_stream_packetout(stream_in, &packet);
	ogg_stream_packetin(stream_out, &packet);

	afhi->header_len = 0;
	while (ogg_stream_flush(stream_out, &page))
		afhi->header_len += page.body_len + page.header_len;
	PARA_DEBUG_LOG("header_len = %d\n", afhi->header_len);
	afhi->header_offset = 0;
	ret = 1;
err2:
	ogg_stream_destroy(stream_in);
	ogg_stream_destroy(stream_out);
err1:
	ogg_sync_destroy(sync_in);
	vorbis_info_clear(&vi);
	vorbis_comment_clear(&vc);
	return ret;
}

/*
 * Alloc and fill array table of byte offsets. chunk_table[i] is the
 * offset in the current input file at which the sample containing time i *
 * CHUNK_TIME begins. Always successful.
 */
static long unsigned ogg_compute_chunk_table(OggVorbis_File *of,
	struct afh_info *afhi, long unsigned time_total)
{
	int i, ret, num;
	ssize_t max_chunk_len, pos = 0, min = 0, old_pos;
	long unsigned num_chunks;

	old_pos = 0;
	ret = 0;
	num = time_total / chunk_time + 3;
	PARA_DEBUG_LOG("chunk time: %g allocating %d chunk pointers\n",
		chunk_time, num);
	afhi->chunk_table = para_malloc((num + 1) * sizeof(size_t));
	afhi->chunk_table[0] = 0;
	max_chunk_len = 0;
	for (i = 1; ret <= num; i++) {
		ogg_int64_t diff;
		ret = ov_time_seek(of, i * chunk_time);
		if (ret)
			break;
		pos = ov_raw_tell(of);
		diff = pos - old_pos;
		max_chunk_len = PARA_MAX(max_chunk_len, diff);
		min = (i == 1)? diff : PARA_MIN(min, diff);
		afhi->chunk_table[i] = pos;
//		if (i < 11 || !((i - 1) % 1000)|| i > num - 11)
//			PARA_DEBUG_LOG("chunk #%d: %g secs, pos: %zd, "
//				"size: %zd\n", i - 1,
//				i * chunk_time, pos, pos - old_pos);
		old_pos = pos;
	}
	num_chunks = i - 1;
//fi->chunk_table[i] = pos;
	PARA_DEBUG_LOG("%lu chunks (%fs), max chunk: %zd, min chunk: %zd\n",
		num_chunks, chunk_time, max_chunk_len, min);
	return num_chunks;
}

/*
 * Init oggvorbis file and write some tech data to given pointers.
 */
static int ogg_get_file_info(char *map, size_t numbytes,
		struct afh_info *afhi)
{
	int ret;
	vorbis_info *vi;
	OggVorbis_File of;
	const ov_callbacks ovc = {
		.read_func = cb_read,
		.seek_func = cb_seek,
		.close_func = cb_close,
		.tell_func = cb_tell
	};
	struct ogg_datasource ods = {.map = map, .numbytes = numbytes, .fpos = 0};

	ret = ogg_compute_header_len(map, numbytes, afhi);
	if (ret < 0)
		return ret;
	ret = ogg_open_callbacks(&ods, &of, ovc);
	if (ret < 0)
		goto err;
	ret = -E_OGG_INFO;
	vi = ov_info(&of, 0);
	if (!vi)
		goto err;
	afhi->seconds_total = ov_time_total(&of, -1);
	afhi->frequency = vi->rate;
	afhi->bitrate = ov_bitrate(&of, 0) / 1000;
	afhi->channels = vi->channels;
	afhi->chunks_total = ogg_compute_chunk_table(&of, afhi, afhi->seconds_total);
	afhi->info_string[0] = '\0';
	sprintf(afhi->info_string, "%s:\n%s:\n%s:\n",
		status_item_list[SI_AUDIO_FILE_INFO],
		status_item_list[SI_TAGINFO1],
		status_item_list[SI_TAGINFO2]
	);
	afhi->chunk_tv.tv_sec = 0;
	afhi->chunk_tv.tv_usec = 250 * 1000;
	tv_scale(10 / afhi->channels, &afhi->chunk_tv, &afhi->eof_tv);
	ret = 1;
err:
	ov_clear(&of); /* keeps the file open */
	return ret;
}

static const char* ogg_suffixes[] = {"ogg", NULL};

/**
 * the init function of the ogg vorbis audio format handler
 *
 * \param afh pointer to the struct to initialize
 */
void ogg_init(struct audio_format_handler *afh)
{
	afh->get_file_info = ogg_get_file_info,
	afh->suffixes = ogg_suffixes;
}
